use std::cmp::min;
use std::ops::{Index, IndexMut, Mul};

use glc::{matrix::Angle, vector::Vector2f};
use serde::{Deserialize, Serialize};
use specs::{Component, VecStorage};

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Ship {
    pub type_: Type,
    pub agility: Angle<f32>,
    pub target_pos: Vector2f,
    pub target_dir: Vector2f,
}

#[derive(Copy, Clone, Debug, Default)]
pub struct Count {
    pub fighter: usize,
    pub bomber: usize,
    pub transporter: usize,
}

#[derive(Copy, Clone, Debug, Serialize, Deserialize)]
pub enum Type {
    Fighter,
    Bomber,
    Transporter,
}

impl Component for Ship {
    type Storage = VecStorage<Self>;
}

impl Count {
    pub fn all() -> Self {
        Self {
            fighter: usize::MAX,
            bomber: usize::MAX,
            transporter: usize::MAX,
        }
    }

    pub fn none() -> Self {
        Self {
            fighter: 0,
            bomber: 0,
            transporter: 0,
        }
    }

    pub fn total(&self) -> usize {
        self.fighter
            .saturating_add(self.bomber)
            .saturating_add(self.transporter)
    }

    pub fn merge(&self, other: &Self) -> Self {
        Self {
            fighter: min(self.fighter, other.fighter),
            bomber: min(self.bomber, other.bomber),
            transporter: min(self.transporter, other.transporter),
        }
    }
}

impl Index<usize> for Count {
    type Output = usize;

    fn index(&self, index: usize) -> &Self::Output {
        match index {
            0 => &self.fighter,
            1 => &self.bomber,
            2 => &self.transporter,
            x => panic!("Invalid ship count index: {}", x),
        }
    }
}

impl IndexMut<usize> for Count {
    fn index_mut(&mut self, index: usize) -> &mut Self::Output {
        match index {
            0 => &mut self.fighter,
            1 => &mut self.bomber,
            2 => &mut self.transporter,
            x => panic!("Invalid ship count index: {}", x),
        }
    }
}

impl Index<Type> for Count {
    type Output = usize;

    fn index(&self, index: Type) -> &Self::Output {
        match index {
            Type::Fighter => &self.fighter,
            Type::Bomber => &self.bomber,
            Type::Transporter => &self.transporter,
        }
    }
}

impl IndexMut<Type> for Count {
    fn index_mut(&mut self, index: Type) -> &mut Self::Output {
        match index {
            Type::Fighter => &mut self.fighter,
            Type::Bomber => &mut self.bomber,
            Type::Transporter => &mut self.transporter,
        }
    }
}

impl Mul<f32> for Count {
    type Output = Count;

    #[allow(unused_assignments)]
    fn mul(self, rhs: f32) -> Self::Output {
        let expected = self.total() as f32;
        let expected = (rhs * expected).ceil() as usize;

        let mut fighter = (rhs * self.fighter as f32) as usize;
        let mut bomber = (rhs * self.bomber as f32) as usize;
        let mut transporter = (rhs * self.transporter as f32) as usize;

        let mut actual = fighter.saturating_add(bomber).saturating_add(transporter);

        if actual < expected && fighter < self.fighter {
            fighter += 1;
            actual += 1;
        }

        if actual < expected && bomber < self.bomber {
            bomber += 1;
            actual += 1;
        }

        if actual < expected && transporter < self.transporter {
            transporter += 1;
            actual += 1;
        }

        Count {
            fighter,
            bomber,
            transporter,
        }
    }
}
